import time

from pymodbus.client.serial import ModbusSerialClient
from pymodbus.exceptions import ModbusException
from pymodbus.pdu import ModbusExceptions, ExceptionResponse
from pymodbus.register_read_message import ReadInputRegistersResponse
from pymodbus.register_write_message import WriteSingleRegisterResponse
from pymodbus.bit_write_message import WriteSingleCoilResponse
from pymodbus.bit_read_message import ReadCoilsResponse

from edge import eventlog
from edge.types.exceptions import InternalError, ToutError


class PsuctlFirmware:
    _SLAVE_ADDRESS = 0x30
    _RX_TO_TX_TOUT_SEC = 0.025

    def __init__(self, endpoint: str) -> None:
        self._modbus = ModbusSerialClient(
            method="rtu", port=endpoint, baudrate=19200,
            bytesize=8, stopbits=1, parity='N')
        self._base_addr = 0
        self._version = None
        self._version = self.version

        if self._version in ['v2.0']:
            self._base_addr = 0x2000

    def read_ro_reg(self, address: int, ack_tout_sec: float) -> int:
        """
        :raise InternalError:
        """
        self._modbus.params.timeout = ack_tout_sec
        try:
            with self._modbus as modbus:
                reply = modbus.read_input_registers(address + self._base_addr, count=1, unit=PsuctlFirmware._SLAVE_ADDRESS)

            time.sleep(self._RX_TO_TX_TOUT_SEC)
            if reply is None:
                eventlog.log_w(f'fail: read_input_registers (address:{address + self._base_addr}, response:None)')
                raise ToutError

            elif isinstance(reply, ReadInputRegistersResponse):
                reg0 = reply.registers[0]

                # Fix byte-order. Firmwares v1.1, v1.2, v1.3 and v1.4 responds
                # in little-endian, instead of big-endian.
                value = ((reg0 >> 8) & 0xFF) | ((reg0 << 8) & 0xFF00)

            elif isinstance(reply, ReadCoilsResponse):
                # Firmware v1.1 responds with a ReadCoilsResponse for a ReadInputRegistersRequest
                #
                # Fix byte-order. Firmwares v1.1, v1.2, v1.3 and v1.4 responds
                # in little-endian, instead of big-endian.
                value = int(''.join(reversed(['1' if bit else '0' for bit in reply.bits])), 2)

            elif isinstance(reply, ExceptionResponse):
                if reply.exception_code == ModbusExceptions.IllegalAddress:
                    eventlog.log_e(f'fail: read_input_registers (address:{address + self._base_addr}, unsupported register)')
                else:
                    eventlog.log_e(f'fail: read_input_registers (address:{address + self._base_addr}, exception:{reply.exception_code})')

                raise InternalError

            else:
                eventlog.log_e(f'fail: read_input_registers (address:{address + self._base_addr}, response:<{reply}>)')
                raise InternalError

            eventlog.log_d(f'done: read_input_registers (address:{address + self._base_addr}, value:{value})')
            return value

        except ModbusException as x:
            eventlog.log_e(f'fail: read_input_registers ({x})')
            raise InternalError

    def write_rw_reg(self, address: int, int16_value: int, ack_tout_sec: float) -> int:
        """
        :returns:
            int16_value
        :raise InternalError:
        """
        eventlog.log_d(f'exec: write_register (address: {hex(address + self._base_addr)}, value{hex(int16_value)})')

        self._modbus.params.timeout = ack_tout_sec
        try:
            with self._modbus as modbus:
                reply = modbus.write_register(address + self._base_addr, int16_value, unit=PsuctlFirmware._SLAVE_ADDRESS)

            time.sleep(PsuctlFirmware._RX_TO_TX_TOUT_SEC)
            if reply is None:
                eventlog.log_w(f'fail: write_register timeout (tout-sec: {ack_tout_sec})')
                raise InternalError

            elif not isinstance(reply, WriteSingleRegisterResponse):
                eventlog.log_w(f'fail: write_register (invalid response <{reply}>)')
                raise InternalError

            elif reply.value != int16_value:
                eventlog.log_w('fail: write_register (address:{}, query-value:{}, reply-value:{})'
                               .format(hex(address + self._base_addr), hex(int16_value), hex(reply.value)))

            else:
                eventlog.log_d(f'done: write_register (address:{hex(address + self._base_addr)}, value:{hex(int16_value)})')

            return reply.value

        except ModbusException as x:
            eventlog.log_e('fail: write_register ({})'.format(str(x)))
            raise InternalError

    def write_rw_bit(self, address: int, int1_value: bool | int, ack_tout_sec: float) -> int:
        """
        :returns:
            int1_value

        :raise InternalError:
        """
        self._modbus.params.timeout = ack_tout_sec
        try:
            with self._modbus as modbus:
                reply = modbus.write_coil(address + self._base_addr, int1_value, unit=PsuctlFirmware._SLAVE_ADDRESS)

            time.sleep(PsuctlFirmware._RX_TO_TX_TOUT_SEC)
            if reply is None:
                eventlog.log_w(f'fail: write_coil timeout (tout-sec: {ack_tout_sec})')
                raise InternalError

            elif not isinstance(reply, WriteSingleCoilResponse):
                eventlog.log_w(f'fail: write_coil (invalid response <{str(reply)}>')
                raise InternalError

            elif reply.value != int1_value:
                eventlog.log_w('fail: write_coil (address:{}, query-value:{}, reply-value:{})'
                               .format(hex(address + self._base_addr), hex(int1_value), hex(reply.value)))

            else:
                eventlog.log_d(f'done: write_coil (address:{hex(address + self._base_addr)}, value:{hex(int1_value)})')

            return reply.value

        except ModbusException as x:
            eventlog.log_e(f'fail: write_coil ({x})')
            raise InternalError

    @property
    def version(self) -> str:
        if self._version is None:
            try:
                # Register 0xFFFF contains the device version regardless of
                # the actual device implementation (this address should not be rebased).
                value = self.read_ro_reg(address=(0xFFFF - self._base_addr), ack_tout_sec=0.25) or 0x10

            except InternalError:
                eventlog.log_e('fail: read_ro_reg')
                raise

            # map 0x14 to 'v1.4'
            ver_major = (value // 16) % 16
            ver_minor = value % 16
            self._version = f'v{ver_major}.{ver_minor}'

        return self._version
